from typing import TYPE_CHECKING, Callable, TypeVar
if TYPE_CHECKING:
    from sodiumfrp.transaction import Transaction

T = TypeVar("T")

Handler = Callable[[T], None]
TransactionHandler = Callable[["Transaction", T], None]



from pathlib import Path
import sys
from threading import Thread
from time import sleep
import tracemalloc as tm

sys.path.append(str(Path(__file__).parent.parent))
from sodiumfrp import Cell, StreamSink

def main() -> None:

    def thread() -> None:
        tm.start()
        while True:
            current, peak = tm.get_traced_memory()
            print(f"memory current: {current}, peak: {peak}")
            sleep(5)
    Thread(target=thread).start()

    et: StreamSink[int] = StreamSink()
    t: Cell[int] = et.hold(0)
    eChange: StreamSink[int] = StreamSink()
    oout = eChange.map(lambda x: t).hold(t)
    out = Cell.switch_cell(oout)
    l = out.listen(lambda tt: None)
    for i in range(1000000000):
        eChange.send(i)
    l.unlisten()

if __name__ == "__main__":
    main()



from typing import Callable, Generic, TypeVar

A = TypeVar("A")
B = TypeVar("B")

class Lazy(Generic[A]):
    """
    A representation for a value that may not be available until the current
    transaction is closed.
    """

    def __init__(self, f: Callable[[], A]) -> None:
        self._f = f

    def get(self) -> A:
        """
        Get the value if available, throwing an exception if not.
        In the general case this should only be used in subsequent
        transactions to when the Lazy was obtained.
        """
        return self._f()

    def lift(self, f: Callable[...,B], *others: "Lazy") -> "Lazy[B]":
        """
        Lift a function into lazy values, so the returned Lazy reflects
        the value of the function applied to the input Lazys' values.
        """
        def helper() -> B:
            other_values = map(lambda x: x.get(), others)
            return f(self.get(), *other_values)
        return Lazy(helper)



from typing import List

from sodiumfrp import operational
from sodiumfrp import StreamSink, Transaction

def test_base_send1() -> None:
    s: StreamSink[str] = Transaction.run(StreamSink)
    out: List[str] = []
    l = Transaction.run(lambda: s.listen(out.append))
    Transaction.run(lambda: s.send("a"))
    Transaction.run(lambda: s.send("b"))
    l.unlisten()
    assert ["a", "b"] == out

def test_operational_split() -> None:
    a: StreamSink[List[str]] = Transaction.run(StreamSink)
    b = Transaction.run(lambda: operational.split(a))
    b_0: List[str] = []
    b_0_l = Transaction.run(lambda: b.listen(b_0.append))
    Transaction.run(lambda: a.send(["a", "b"]))
    b_0_l.unlisten()
    assert ["a", "b"] == b_0

def test_operational_defer1() -> None:
    a: StreamSink[str] = Transaction.run(StreamSink)
    b = Transaction.run(lambda: operational.defer(a))
    b_0: List[str] = []
    b_0_l = Transaction.run(lambda: b.listen(b_0.append))
    Transaction.run(lambda: a.send("a"))
    b_0_l.unlisten()
    assert ["a"] == b_0
    b_1: List[str] = []
    b_1_l = Transaction.run(lambda: b.listen(b_1.append))
    Transaction.run(lambda: a.send("b"))
    b_1_l.unlisten()
    assert ["b"] == b_1

def test_operational_defer2() -> None:
    a: StreamSink[str] = Transaction.run(StreamSink)
    b: StreamSink[str] = Transaction.run(StreamSink)
    c = Transaction.run(lambda: operational.defer(a).or_else(b))
    c_0: List[str] = []
    c_0_l = Transaction.run(lambda: c.listen(c_0.append))
    Transaction.run(lambda: a.send("a"))
    c_0_l.unlisten()
    assert ["a"] == c_0
    c_1: List[str] = []
    c_1_l = Transaction.run(lambda: c.listen(c_1.append))
    def transaction() -> None:
        a.send("b")
        b.send("B")
    Transaction.run(transaction)
    c_1_l.unlisten()
    assert ["B", "b"] == c_1

def test_stream_or_else1() -> None:
    a: StreamSink[int] = Transaction.run(StreamSink)
    b: StreamSink[int] = Transaction.run(StreamSink)
    c = Transaction.run(lambda: a.or_else(b))
    c_0: List[int] = []
    c_0_l = Transaction.run(lambda: c.listen(c_0.append))
    Transaction.run(lambda: a.send(0))
    c_0_l.unlisten()
    assert [0] == c_0
    c_1: List[int] = []
    c_1_l = Transaction.run(lambda: c.listen(c_1.append))
    Transaction.run(lambda: b.send(10))
    c_1_l.unlisten()
    assert [10] == c_1
    c_2: List[int] = []
    c_2_l = Transaction.run(lambda: c.listen(c_2.append))
    def transaction() -> None:
        a.send(2)
        b.send(20)
    Transaction.run(transaction)
    c_2_l.unlisten()
    assert [2] == c_2
    c_3: List[int] = []
    c_3_l = Transaction.run(lambda: c.listen(c_3.append))
    Transaction.run(lambda: b.send(30))
    c_3_l.unlisten()
    assert [30] == c_3

def test_operational_defer_simultaneous() -> None:
    a: StreamSink[str] = Transaction.run(StreamSink)
    b: StreamSink[str] = Transaction.run(StreamSink)
    c = Transaction.run(
        lambda: operational.defer(a).or_else(operational.defer(b)))
    c_0: List[str] = []
    c_0_l = Transaction.run(lambda: c.listen(c_0.append))
    Transaction.run(lambda: b.send("A"))
    c_0_l.unlisten()
    assert ["A"] == c_0
    c_1: List[str] = []
    c_1_l = Transaction.run(lambda: c.listen(c_1.append))
    def transaction() -> None:
        a.send("b")
        b.send("B")
    Transaction.run(transaction)
    c_1_l.unlisten()
    assert ["b"] == c_1



from collections import namedtuple
from typing import Any, List, Optional, Tuple

from sodiumfrp import operational
from sodiumfrp import \
    Cell, \
    CellLoop, \
    CellSink, \
    Listener, \
    Stream, \
    StreamSink, \
    Transaction, \
    Unit, \
    UNIT

def test_hold() -> None:
    e: StreamSink[int] = StreamSink()
    b = e.hold(0)
    out: List[int] = []
    l = operational.updates(b).listen(out.append)
    e.send(2)
    e.send(9)
    l.unlisten()
    assert [2, 9] == out

def test_snapshot() -> None:
    b: CellSink[int] = CellSink(0)
    trigger: StreamSink[int] = StreamSink()
    out: List[str] = []
    l = trigger \
        .snapshot(b, lambda x, y: f"{x} {y}") \
        .listen(out.append)
    trigger.send(100)
    b.send(2)
    trigger.send(200)
    b.send(9)
    b.send(1)
    trigger.send(300)
    l.unlisten()
    assert ["100 0", "200 2", "300 1"] == out

def test_snapshot_default() -> None:
    b: CellSink[int] = CellSink(0)
    trigger: StreamSink[int] = StreamSink()
    out: List[str] = []
    l = trigger \
        .snapshot(b) \
        .listen(out.append)
    trigger.send(100)
    b.send(2)
    trigger.send(200)
    b.send(9)
    b.send(1)
    trigger.send(300)
    l.unlisten()
    assert [0, 2, 1] == out

def test_snapshot_many() -> None:
    b: CellSink[int] = CellSink(0)
    c: CellSink[int] = CellSink(1)
    trigger: StreamSink[int] = StreamSink()
    out: List[str] = []
    l = trigger \
        .snapshot(b, c, lambda a, b, c: f"{a} {b} {c}") \
        .listen(out.append)
    trigger.send(100)
    b.send(2)
    c.send(3)
    trigger.send(200)
    b.send(9)
    b.send(1)
    trigger.send(300)
    l.unlisten()
    assert ["100 0 1", "200 2 3", "300 1 3"] == out

def test_snapshot_many_default() -> None:
    b: CellSink[int] = CellSink(0)
    c: CellSink[int] = CellSink(1)
    trigger: StreamSink[int] = StreamSink()
    out: List[str] = []
    l = trigger \
        .snapshot(b, c) \
        .listen(out.append)
    trigger.send(100)
    c.send(3)
    b.send(2)
    trigger.send(200)
    b.send(9)
    b.send(1)
    trigger.send(300)
    l.unlisten()
    assert [(0,1), (2,3), (1,3)] == out

def test_hold_is_delayed() -> None:
    e: StreamSink[int] = StreamSink()
    h = e.hold(0)
    pair = e.snapshot(h, lambda a, b: f"{a} {b}")
    out: List[str] = []
    l = pair.listen(out.append)
    e.send(2)
    e.send(3)
    l.unlisten()
    assert ["2 0", "3 2"] == out

def test_sample() -> None:
    b: CellSink = CellSink(0)
    out: List[int] = []
    out.append(b.sample())
    b.send(3)
    out.append(b.sample())
    b.send(42)
    out.append(b.sample())
    assert [0, 3, 42] == out

def test_map() -> None:
    b = CellSink(6)
    out: List[str] = []
    l = b.map(str).listen(out.append)
    b.send(8)
    l.unlisten()
    assert ["6", "8"] == out

def test_map_late_listen() -> None:
    b = CellSink(6)
    out: List[str] = []
    bm = b.map(str)
    b.send(2)
    l = bm.listen(out.append)
    b.send(8)
    l.unlisten()
    assert ["2", "8"] == out

def test_starmap() -> None:
    b = CellSink((1, 2))
    out: List[str] = []
    l = b.starmap(lambda x, y: f"{x} {y}").listen(out.append)
    b.send((3, 4))
    l.unlisten()
    assert ["1 2", "3 4"] == out

def test_apply() -> None:
    bf = CellSink(lambda b: f"1 {b}")
    ba = CellSink(5)
    out: List[str] = []
    l = Cell.apply(bf, ba).listen(out.append)
    bf.send(lambda b: f"12 {b}")
    ba.send(6)
    l.unlisten()
    assert ["1 5", "12 5", "12 6"] == out

def test_lift() -> None:
    a = CellSink(1)
    b = CellSink(5)
    out: List[str] = []
    l = a.lift(b, lambda x, y: f"{x} {y}").listen(out.append)
    a.send(12)
    b.send(6)
    l.unlisten()
    assert ["1 5", "12 5", "12 6"] == out

def test_lift_glitch() -> None:
    a = CellSink(1)
    a3 = a.map(lambda x: x * 3)
    a5 = a.map(lambda x: x * 5)
    b = a3.lift(a5, lambda x, y: f"{x} {y}")
    out: List[str] = []
    l = b.listen(out.append)
    a.send(2)
    l.unlisten()
    assert ["3 5", "6 10"] == out

def test_lift_from_simultaneous() -> None:
    def transaction() -> Tuple[Cell,Cell]:
        b1 = CellSink(3)
        b2 = CellSink(5)
        b2.send(7)
        return (b1, b2)
    b1, b2 = Transaction.run(transaction)
    out: List[int] = []
    l = b1.lift(b2, lambda x, y: x + y).listen(out.append)
    l.unlisten()
    assert [10] == out

def test_listen() -> None:
    b = CellSink(9)
    out: List[int] = []
    l = b.listen(out.append)
    b.send(2)
    b.send(7)
    l.unlisten()
    assert [9, 2, 7] == out

def test_constant() -> None:
    b = Cell.constant(12)
    out: List[int] = []
    l = b.listen(out.append)
    l.unlisten()
    assert [12] == out

def test_switch_cell() -> None:
    SB = namedtuple("SB", ["a", "b", "sw"])
    esb: StreamSink[SB] = StreamSink()
    # Split each field out of SB so we can update multiple behaviours in a
    # single transaction.
    ba = esb.map(lambda s: s.a).filter(not_none).hold("A")
    bb = esb.map(lambda s: s.b).filter(not_none).hold("a")
    bsw = esb.map(lambda s: s.sw).filter(not_none).hold(ba)
    bo = Cell.switch_cell(bsw)
    out: List[str] = []
    l = bo.listen(out.append)
    esb.send(SB("B", "b", None))
    esb.send(SB("C", "c", bb))
    esb.send(SB("D", "d", None))
    esb.send(SB("E", "e", ba))
    esb.send(SB("F", "f", None))
    esb.send(SB(None, None, bb))
    esb.send(SB(None, None, ba))
    esb.send(SB("G", "g", bb))
    esb.send(SB("H", "h", ba))
    esb.send(SB("I", "i", ba))
    l.unlisten()
    assert ["A", "B", "c", "d", "E", "F", "f", "F", "g", "H", "I"] == out

def test_switch_stream() -> None:
    SE = namedtuple("SE", ["a", "b", "sw"])
    ese: StreamSink[SE] = StreamSink()
    ea = ese.map(lambda s: s.a)
    eb = ese.map(lambda s: s.b)
    bsw = ese.map(lambda s: s.sw).filter(not_none).hold(ea)
    out: List[str] = []
    eo = Cell.switch_stream(bsw)
    l = eo.listen(out.append)
    ese.send(SE('A', 'a', None))
    ese.send(SE('B', 'b', None))
    ese.send(SE('C', 'c', eb))
    ese.send(SE('D', 'd', None))
    ese.send(SE('E', 'e', ea))
    ese.send(SE('F', 'f', None))
    ese.send(SE('G', 'g', eb))
    ese.send(SE('H', 'h', ea))
    ese.send(SE('I', 'i', ea))
    l.unlisten()
    assert ['A', 'B', 'C', 'd', 'e', 'F', 'G', 'h', 'I'] == out

def test_switch_stream_simultaneous() -> None:

    class SS2:
        def __init__(self) -> None:
            self.s: StreamSink[int] = StreamSink()

    ss1 = SS2()
    css = CellSink(ss1)
    so = Cell.switch_stream(css.map(lambda b: b.s))
    out: List[int] = []
    l = so.listen(out.append)
    ss3 = SS2()
    ss4 = SS2()
    ss2 = SS2()
    ss1.s.send(0)
    ss1.s.send(1)
    ss1.s.send(2)
    css.send(ss2)
    ss1.s.send(7)
    ss2.s.send(3)
    ss2.s.send(4)
    ss3.s.send(2)
    css.send(ss3)
    ss3.s.send(5)
    ss3.s.send(6)
    ss3.s.send(7)
    def transaction() -> None:
        ss3.s.send(8)
        css.send(ss4)
        ss4.s.send(2)
    Transaction.run(transaction)
    ss4.s.send(9)
    l.unlisten()
    assert [0, 1, 2, 3, 4, 5, 6, 7, 8, 9] == out

def test_loop_cell() -> None:
    sa: StreamSink[int] = StreamSink()
    def transaction() -> Cell[int]:
        sum_: CellLoop[int] = CellLoop()
        sum_out_ = sa.snapshot(sum_, lambda x, y: x + y).hold(0)
        sum_.loop(sum_out_)
        return sum_out_
    sum_out = Transaction.run(transaction)
    out: List[int] = []
    l = sum_out.listen(out.append)
    sa.send(2)
    sa.send(3)
    sa.send(1)
    l.unlisten()
    assert [0, 2, 5, 6] == out
    assert 6 == sum_out.sample()

def test_loop_value_snapshot() -> None:
    out: List[str] = []
    def transaction() -> Listener:
        a = Cell.constant("lettuce")
        b: CellLoop[str] = CellLoop()
        eSnap = operational.value(a).snapshot(b, lambda x, y: f"{x} {y}")
        b.loop(Cell.constant("cheese"))
        return eSnap.listen(out.append)
    l = Transaction.run(transaction)
    l.unlisten()
    assert ["lettuce cheese"] == out

def test_loop_value_hold() -> None:
    out: List[str] = []
    def transaction() -> Cell[str]:
        a: CellLoop[str] = CellLoop()
        value_ = operational.value(a).hold("onion")
        a.loop(Cell.constant("cheese"))
        return value_
    value = Transaction.run(transaction)
    eTick: StreamSink[Unit] = StreamSink()
    l = eTick.snapshot(value).listen(out.append)
    eTick.send(UNIT)
    l.unlisten()
    assert ["cheese"] == out

def test_lift_loop() -> None:
    out: List[str] = []
    b = CellSink("kettle")
    def transaction() -> Cell[str]:
        a: CellLoop[str] = CellLoop()
        c_ = a.lift(b, lambda x, y: f"{x} {y}")
        a.loop(Cell.constant("tea"))
        return c_
    c = Transaction.run(transaction)
    l = c.listen(out.append)
    b.send("caddy")
    l.unlisten()
    assert ["tea kettle", "tea caddy"] == out

def test_switch_and_defer() -> None:
    out: List[str] = []
    si: StreamSink[int] = StreamSink()
    l = Cell.switch_stream(
            si.map(
                lambda i: operational.defer(
                    operational.value(Cell.constant(f"A{i}"))
                )
            ).hold(Stream.never())
        ).listen(out.append)
    si.send(2)
    si.send(4)
    l.unlisten()
    assert ["A2", "A4"] == out

def test_calm() -> None:
    b: CellSink[int] = CellSink(42)
    out: List[int] = []
    l = b.calm().listen(out.append)
    ins = [
        1, 1,
        2, 2, 2,
        1,
        3, 3,
        1, 1, 1,
        2, 2, 2, 2,
        3,
    ]
    for i in ins:
        b.send(i)
    l.unlisten()
    assert [42, 1, 2, 1, 3, 1, 2, 3] == out

def test_calm_repeat_inital_value() -> None:
    b: CellSink[int] = CellSink(1)
    out: List[int] = []
    l = b.calm().listen(out.append)
    ins = [
        1, 1,
        2, 2, 2,
        1,
        3, 3,
        1, 1, 1,
        2, 2, 2, 2,
        3,
    ]
    for i in ins:
        b.send(i)
    l.unlisten()
    assert [1, 2, 1, 3, 1, 2, 3] == out

def test_calm_merge() -> None:
    out: List = []

    def init() -> Tuple:
        b1: CellSink[int] = CellSink(1)
        b2 = b1.calm()
        e1 = operational.value(b1)
        e2 = operational.value(b2)
        lm = e1.merge_with(e2, lambda x, y: (x, y)).listen(out.append)
        return b1, lm

    b1, lm = Transaction.run(init)
    ins = [
        1, 1,
        2, 2, 2,
        1,
        3, 3,
        1, 1, 1,
        2, 2, 2, 2,
        3,
    ]
    for i in ins:
        b1.send(i)
    lm.unlisten()
    assert [
            (1, 1), 1, 1,
            (2, 2), 2, 2,
            (1, 1),
            (3, 3), 3,
            (1, 1), 1, 1,
            (2, 2), 2, 2, 2,
            (3, 3),
        ] == out

def not_none(x: Optional[Any]) -> bool:
    return x is not None



""" Operational primitives that must be used with care. """

from typing import Iterable, TypeVar

from sodiumfrp.primitives import Cell, Stream, StreamWithSend
from sodiumfrp.transaction import Transaction

A = TypeVar("A")

def updates(c: Cell[A]) -> Stream[A]:
    """
    A stream that gives the updates/steps for a :class:`Cell`.

    This is an **operational** primitive, which is not part of the main Sodium
    API. It breaks the property of non-detectability of cell steps/updates.
    The rule with this primitive is that you should only use it in functions
    that do not allow the caller to detect the cell updates.
    """
    return c._updates()


def value(c: Cell[A]) -> Stream[A]:
    """
    A stream that is guaranteed to fire once in the transaction where
    `value()` is invoked, giving the current value of the cell, and thereafter
    behaves like :meth:`updates()`, firing for each update/step of
    the cell's value.

    This is an **operational** primitive, which is not part of the main Sodium
    API. It breaks the property of non-detectability of cell steps/updates.
    The rule with this primitive is that you should only use it in functions
    that do not allow the caller to detect the cell updates.
    """
    return Transaction._apply(c._value_stream)


def defer(s: Stream[A]) -> Stream[A]:
    """
    Push each event onto a new transaction guaranteed to come before
    the next externally initiated transaction. Same as :meth:`split` but it
    works on a single value.
    """
    return split(s.map(lambda a: [a]))


def split(s: Stream[Iterable[A]]) -> Stream[A]:
    """
    Push each event in the list onto a newly created transaction guaranteed
    to come before the next externally initiated transaction. Note that
    the semantics are such that two different invocations of split() can
    put events into the same new transaction, so the resulting stream's
    events could be simultaneous with events output by :meth:`split` or
    :meth:`defer` invoked elsewhere in the code.
    """
    out: StreamWithSend[A] = StreamWithSend()

    def handler(trans: Transaction, as_: Iterable[A]) -> None:
        child_idx = 0
        for a in as_:
            def run(trans2: Transaction, _a: A = a) -> None:
                out._send(trans2, _a)
            trans._post(child_idx, run)
            child_idx += 1

    l1 = s._listen(out._node, handler)
    return out._unsafe_add_cleanup(l1)



from time import sleep
from threading import Thread
from typing import List

from sodiumfrp import Transaction

def test_post() -> None:
    out: List[int] = []
    Transaction.post(lambda: out.append(42))
    assert [42] == out

def test_post_overlapping() -> None:
    out: List[int] = []
    def thread() -> None:
        sleep(0.02)
        Transaction.post(lambda: out.append(42))
    Thread(target=thread).start()
    Transaction.run(lambda: sleep(0.04))
    assert [42] == out



from typing import List, Tuple

from sodiumfrp import operational
from sodiumfrp import CellSink, Stream, StreamLoop, StreamSink, Transaction

def test_send_stream() -> None:
    e: StreamSink[int] = StreamSink()
    out: List[int] = []
    l = e.listen(out.append)
    e.send(5)
    l.unlisten()
    assert [5] == out
    e.send(6)
    assert [5] == out

def test_map() -> None:
    e: StreamSink[int] = StreamSink()
    m = e.map(str)
    out: List[str] = []
    l = m.listen(out.append)
    e.send(5)
    l.unlisten()
    assert ["5"] == out

def test_map_to() -> None:
    e: StreamSink[int] = StreamSink()
    m = e.map_to("fusebox")
    out: List[str] = []
    l = m.listen(out.append)
    e.send(5)
    e.send(6)
    l.unlisten()
    assert ["fusebox", "fusebox"] == out

def test_starmap() -> None:
    e: StreamSink[Tuple] = StreamSink()
    m = e.starmap(lambda x, y: f"{x} {y}")
    out: List[str] = []
    l = m.listen(out.append)
    e.send((1, 2))
    l.unlisten()
    assert ["1 2"] == out

def test_merge_non_simultaneous() -> None:
    e1: StreamSink[int] = StreamSink()
    e2: StreamSink[int] = StreamSink()
    out: List[int] = []
    l = e2.or_else(e1).listen(out.append)
    e1.send(7)
    e2.send(9)
    e1.send(8)
    l.unlisten()
    assert [7, 9, 8] == out

def test_merge_simultaneous() -> None:
    s1: StreamSink[int] = StreamSink(lambda l, r: r)
    s2: StreamSink[int] = StreamSink(lambda l, r: r)
    out: List[int] = []
    l = s2.or_else(s1).listen(out.append)

    def trans1() -> None:
        s1.send(7)
        s2.send(60)
    Transaction.run(trans1)

    def trans2() -> None:
        s1.send(9)
    Transaction.run(trans2)

    def trans3() -> None:
        s1.send(7)
        s1.send(60)
        s2.send(8)
        s2.send(90)
    Transaction.run(trans3)

    def trans4() -> None:
        s2.send(8)
        s2.send(90)
        s1.send(7)
        s1.send(60)
    Transaction.run(trans4)

    def trans5() -> None:
        s2.send(8)
        s1.send(7)
        s2.send(90)
        s1.send(60)
    Transaction.run(trans5)

    l.unlisten()
    assert [60, 9, 90, 90, 90] == out

def test_coalesce() -> None:
    s: StreamSink[int] = StreamSink(lambda a, b: a + b)
    out: List[int] = []
    l = s.listen(out.append)

    def trans1() -> None:
        s.send(2)
    Transaction.run(trans1)

    def trans2() -> None:
        s.send(8)
        s.send(40)
    Transaction.run(trans2)

    l.unlisten()
    assert [2, 48] == out

def test_merge_multiple() -> None:
    s1: StreamSink[int] = StreamSink()
    s2: StreamSink[int] = StreamSink()
    s3: StreamSink[int] = StreamSink()
    out: List[int] = []
    l = Stream.merge(lambda a, b: a + b, s1, s2, s3).listen(out.append)

    def trans1() -> None:
        s3.send(7)
        s1.send(1)
    Transaction.run(trans1)

    def trans2() -> None:
        s1.send(9)
    Transaction.run(trans2)

    def trans3() -> None:
        s2.send(6)
        s3.send(3)
        s1.send(2)
    Transaction.run(trans3)

    def trans4() -> None:
        s2.send(9)
        s1.send(1)
    Transaction.run(trans4)

    def trans5() -> None:
        s3.send(5)
        s1.send(9)
        s2.send(3)
    Transaction.run(trans5)

    l.unlisten()
    assert [8, 9, 11, 10, 17] == out

def test_or_else_multiple() -> None:
    s1: StreamSink[int] = StreamSink()
    s2: StreamSink[int] = StreamSink()
    s3: StreamSink[int] = StreamSink()
    out: List[int] = []
    l = Stream.or_else_(s1, s2, s3).listen(out.append)

    def trans1() -> None:
        s3.send(7)
        s1.send(1)
    Transaction.run(trans1)

    def trans2() -> None:
        s3.send(9)
    Transaction.run(trans2)

    def trans3() -> None:
        s2.send(6)
        s3.send(3)
        s1.send(2)
    Transaction.run(trans3)

    def trans4() -> None:
        s2.send(9)
        s1.send(1)
    Transaction.run(trans4)

    def trans5() -> None:
        s3.send(5)
        s1.send(9)
        s2.send(3)
    Transaction.run(trans5)

    l.unlisten()
    assert [1, 9, 2, 1, 9] == out

def test_filter() -> None:
    e: StreamSink[str] = StreamSink()
    out: List[str] = []
    l = e.filter(str.isupper).listen(out.append)
    e.send("H")
    e.send("o")
    e.send("I")
    l.unlisten()
    assert ["H", "I"] == out

def test_gate() -> None:
    ec: StreamSink[str] = StreamSink()
    epred = CellSink(True)
    out: List[str] = []
    l = ec.gate(epred).listen(out.append)
    ec.send("H")
    epred.send(False)
    ec.send("O")
    epred.send(True)
    ec.send("I")
    l.unlisten()
    assert ["H", "I"] == out

def test_gate_none() -> None:
    ec: StreamSink[None] = StreamSink()
    epred = CellSink(True)
    out: List[None] = []
    l = ec.gate(epred).listen(out.append)
    ec.send(None)
    epred.send(False)
    ec.send(None)
    epred.send(True)
    ec.send(None)
    l.unlisten()
    assert [None, None] == out

def test_loop_stream() -> None:
    ea: StreamSink[int] = StreamSink()
    def transaction() -> Stream[int]:
        eb: StreamLoop[int] = StreamLoop()
        ec_ = ea.map(lambda x: x % 10).merge_with(eb, lambda x, y: x + y)
        eb_out = ea.map(lambda x: x // 10).filter(lambda x: x != 0)
        eb.loop(eb_out)
        return ec_
    ec = Transaction.run(transaction)
    out: List[int] = []
    l = ec.listen(out.append)
    ea.send(2)
    ea.send(52)
    l.unlisten()
    assert [2, 7] == out

def test_collect() -> None:
    ea: StreamSink[int] = StreamSink()
    out: List[int] = []
    sum_ = ea.collect(0, lambda a, s: (a + s + 100, a + s))
    l = sum_.listen(out.append)
    ea.send(5)
    ea.send(7)
    ea.send(1)
    ea.send(2)
    ea.send(3)
    l.unlisten()
    assert [105, 112, 113, 115 ,118] == out

def test_accum() -> None:
    ea: StreamSink[int] = StreamSink()
    out: List[int] = []
    sum_ = ea.accum(100, lambda a, s: a + s)
    l = sum_.listen(out.append)
    ea.send(5)
    ea.send(7)
    ea.send(1)
    ea.send(2)
    ea.send(3)
    l.unlisten()
    assert [100, 105, 112, 113, 115, 118] == out

def test_once() -> None:
    e: StreamSink[str] = StreamSink()
    out: List[str] = []
    l = e.once().listen(out.append)
    e.send("A")
    e.send("B")
    e.send("C")
    l.unlisten()
    assert ["A"] == out

def test_defer() -> None:
    e: StreamSink[str] = StreamSink()
    b = e.hold(" ")
    out: List[str] = []
    l = operational.defer(e).snapshot(b).listen(out.append)
    e.send("C")
    e.send("B")
    e.send("A")
    l.unlisten()
    assert ["C", "B", "A"] == out

def test_calm() -> None:
    e: StreamSink[int] = StreamSink()
    out: List[int] = []
    l = e.calm().listen(out.append)
    ins = [
        1, 1,
        2, 2, 2,
        1,
        3, 3,
        1, 1, 1,
        2, 2, 2, 2,
        3,
    ]
    for i in ins:
        e.send(i)
    l.unlisten()
    assert [1, 2, 1, 3, 1, 2, 3] == out



from .primitives import \
    Cell, \
    CellLoop, \
    CellSink, \
    Stream, \
    StreamLoop, \
    StreamSink
from .listener import Listener
from .transaction import Transaction
from .unit import Unit, UNIT



from time import sleep, monotonic
from typing import List, Optional

from sodiumfrp import Cell, CellLoop, CellSink, Listener, Stream, Transaction
from sodiumfrp.time import TimerSystem

def test_time_sample() -> None:
    with TimerSystem() as ts:
        t1 = Transaction.run(ts.time_ms().sample)
        sleep(0.1)
        t2 = Transaction.run(ts.time_ms().sample)
        delta = t2 - t1
        # This test may fail occasionally, because there is no way to measure
        # time precisely
        assert abs(delta - 100) < 5

def test_at() -> None:
    with TimerSystem() as ts:
        time_ = Transaction.run(ts.time_ms().sample)
        alarm = ts.at(Cell.constant(time_ + 100))
        out: List[int] = []
        alarm.listen(lambda t: out.append(t - time_))
        sleep(0.11)
        assert [100] == out

def test_merge_non_simultaneous() -> None:
    with TimerSystem() as ts:
        out: List[int] = []
        now = Transaction.run(ts.time_ms().sample)
        a1 = ts.at(Cell.constant(now + 100))
        a2 = ts.at(Cell.constant(now + 101))
        m = Stream.or_else_(
            a1.map_to(1),
            a2.map_to(2),
        )
        m.listen(out.append)
        sleep(0.11)
        assert [1, 2] == out

def test_merge_simultaneous() -> None:
    with TimerSystem() as ts:
        out: List[int] = []
        now = Transaction.run(ts.time_ms().sample)
        a1 = ts.at(Cell.constant(now + 100))
        a2 = ts.at(Cell.constant(now + 100))
        m = a1.or_else(a2)
        m.listen(out.append)
        sleep(0.11)
        assert 1 == len(out)

def test_periodic() -> None:
    with TimerSystem() as ts:
        errors: List[float] = []

        def transaction() -> None:
            t0 = ts.time_ms().sample()

            def update_alarm(t: int) -> Optional[int]:
                if ts.time_ms().sample() < (t0 + 110):
                    return t + 1
                return None

            alarm_time: CellLoop[Optional[int]] = CellLoop()
            alarm = ts.at(alarm_time)
            alarm_time.loop(alarm.map(update_alarm).hold(t0 + 10))

            def error(t: int) -> float:
                return abs(monotonic() * 1000 - t)
            alarm.listen(lambda t: errors.append(error(t)))

        Transaction.run(transaction)

        sleep(0.12)

        assert len(errors) == 102
        rmse = pow(sum(map(lambda x: x * x, errors)) / len(errors), 0.5)
        # This test may fail occasionally, because there is no way to measure
        # time precisely
        assert rmse < 1.0

def test_passed_time() -> None:
    with TimerSystem() as ts:
        out: List[int] = []
        def transaction() -> Listener:
            now = ts.time_ms().sample()
            alarm = ts.at(Cell.constant(now))
            return alarm.listen(out.append)
        l = Transaction.run(transaction)
        sleep(0.1)
        l.unlisten()
        assert len(out) == 0

def test_cancel_alarm() -> None:
    with TimerSystem() as ts:
        alarm_time: CellSink[Optional[int]] = CellSink(None)
        out: List[int] = []
        def transaction() -> Listener:
            now = ts.time_ms().sample()
            alarm_time.send(now + 100)
            alarm = ts.at(alarm_time)
            return alarm.listen(out.append)
        l = Transaction.run(transaction)
        sleep(0.055)
        alarm_time.send(None)
        sleep(0.055)
        l.unlisten()
        assert len(out) == 0



from typing import List, Set
import weakref

from sodiumfrp.unit import Unit
from sodiumfrp.typing import TransactionHandler


class Target:

    def __init__(self,
            action: "TransactionHandler[Unit]",
            node: "Node") -> None:
        self.action = weakref.ref(action)
        self.node = node


class Node:

    def __init__(self, rank: int) -> None:
        self._rank = rank
        self._listeners: List[Target] = []

    def _link_to(self,
            action: "TransactionHandler[Unit]",
            target: "Node",
            out_target: List[Target]) -> bool:
        """ Return true if any changes were made. """
        changed = target._ensure_bigger_than(self._rank, set())
        t = Target(action, target)
        self._listeners.append(t)
        out_target[0] = t
        return changed

    def _unlink_to(self, target: Target) -> None:
        self._listeners.remove(target)

    def _ensure_bigger_than(self, limit: int, visited: Set["Node"]) -> bool:
        if self._rank > limit or (self in visited):
            return False

        visited.add(self)
        self._rank = limit + 1
        for listener in self._listeners:
            listener.node._ensure_bigger_than(self._rank, visited)
        visited.remove(self)
        return True

    def __lt__(self, other: "Node") -> bool:
        return self._rank < other._rank

    def __gt__(self, other: "Node") -> bool:
        return self._rank > other._rank


NODE_NULL = Node(9_223_372_036_854_775_807)



from typing import Callable

class Listener:
    """
    A handle for a listener that was registered with :meth:`Cell.listen()
    <sodiumfrp.primitives.Cell.listen>` or :meth:`Stream.listen()
    <sodiumfrp.primitives.Stream.listen>`.
    """

    def __init__(self, unlisten: Callable[["Listener"], None]) -> None:
        self._unlisten = unlisten

    def unlisten(self) -> None:
        """
        Deregister the listener that was registered so it will no longer
        be called back, allowing associated resources to be garbage
        collected.
        """
        self._unlisten(self)

    def append(self, two: "Listener") -> "Listener":
        """
        Combine listeners into one so that invoking :meth:`unlisten` on
        the returned listener will unlisten both the inputs.
        """
        one = self
        def unlisten(_: "Listener") -> None:
            one.unlisten()
            two.unlisten()
        return Listener(unlisten)



import setuptools

def get_long_description():
    with open("README.md", "r") as readme_file:
        return readme_file.read()

setuptools.setup(
    name="sodiumfrp",
    version="1.0.0",
    author="Stephen Blackheath",
    description= \
        "Python implementation of Sodium - "
        "Functional Reactive Programming (FRP) library",
    long_description=get_long_description(),
    long_description_content_type="text/markdown",
    url="https://github.com/mode89/sodium-python",
    packages=setuptools.find_packages(),
    package_data={"sodiumfrp": ["py.typed"]},
    license="BSD-3 License",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6"
)



from heapq import heappush, heappop
from threading import RLock
from typing import Callable, Dict, List, Optional, TypeVar

from sodiumfrp.node import Node
from sodiumfrp.typing import Handler, TransactionHandler

T = TypeVar("T")


class Entry:

    _next_seq: int = 0

    def __init__(self,
            rank: Node,
            action: Handler["Transaction"]) -> None:
        self.rank = rank
        self.action = action
        self.seq = Entry._next_seq
        Entry._next_seq += 1

    def __lt__(self, other: "Entry") -> bool:
        if self.rank < other.rank:
            return True
        elif self.rank > other.rank:
            return False
        # Same rank: preserve chronological sequence.
        return self.seq < other.seq


class Transaction:
    """ Functions for controlling transactions. """

    # Coarse-grained lock that's held during the whole transaction.
    _transaction_lock: RLock = RLock()
    _listeners_lock: RLock = RLock()
    _current_transaction: "Transaction" = None
    in_callback: int = 0
    _on_start_hooks: List[Callable[[], None]] = []
    _running_on_start_hooks: bool = False
    _post_q_lock = RLock()

    def __init__(self) -> None:
        # True if we need to re-generate the priority queue.
        self._to_regen = False
        self._prioritized_q: List[Entry] = []
        self._last_q: List[Callable[[], None]] = []
        self._post_q: Dict[int, Handler["Transaction"]] = None
        self._closed = False

    @staticmethod
    def get_current_transaction() -> "Transaction":
        """ Return the current transaction, or `None` if there isn't one. """
        with Transaction._transaction_lock:
            return Transaction._current_transaction

    @staticmethod
    def run(code: Callable[[], T]) -> T:
        """
        Run the specified code inside a single transaction, with
        the contained code returning a value of the parameter type `T`.

        In most cases this is not needed, because the primitives always
        create their own transaction automatically, but it is needed in some
        circumstances.
        """
        with Transaction._transaction_lock:
            # If we are already inside a transaction (which must be on
            # the same thread otherwise we wouldn't have acquired
            # transactionLock), then keep using that same transaction.
            prev_trans = Transaction._current_transaction
            try:
                Transaction._start_if_necessary()
                return code()
            finally:
                try:
                    if prev_trans is None:
                        Transaction._current_transaction.close()
                finally:
                    Transaction._current_transaction = prev_trans

    @staticmethod
    def _run_handler(code: Handler["Transaction"]) -> None:
        with Transaction._transaction_lock:
            # If we are already inside a transaction (which must be on
            # the same thread otherwise we wouldn't have acquired
            # transactionLock), then keep using that same transaction.
            prev_trans = Transaction._current_transaction
            try:
                Transaction._start_if_necessary()
                code(Transaction._current_transaction)
            finally:
                try:
                    if prev_trans is None:
                        Transaction._current_transaction.close()
                finally:
                    Transaction._current_transaction = prev_trans

    @staticmethod
    def on_start(runnable: Callable[[], None]) -> Callable[[], None]:
        """
        Add a runnable that will be executed whenever a transaction is
        started. That runnable may start transactions itself, which will
        not cause the hooks to be run recursively.

        The main use case of this is the implementation of a time/alarm
        system.

        Returns a callable which, when called, unregisters the hook.
        """
        with Transaction._transaction_lock:
            Transaction._on_start_hooks.append(runnable)
            def unregister_hook() -> None:
                with Transaction._transaction_lock:
                    Transaction._on_start_hooks.remove(runnable)
            return unregister_hook

    @staticmethod
    def _apply(code: Callable[["Transaction"], T]) -> T:
        with Transaction._transaction_lock:
            # If we are already inside a transaction (which must be on
            # the same thread otherwise we wouldn't have acquired
            # transactionLock), then keep using that same transaction.
            prev_trans = Transaction._current_transaction
            try:
                Transaction._start_if_necessary()
                return code(Transaction._current_transaction)
            finally:
                try:
                    if prev_trans is None:
                        Transaction._current_transaction.close()
                finally:
                    Transaction._current_transaction = prev_trans

    @staticmethod
    def _start_if_necessary() -> None:
        if Transaction._current_transaction is None:
            if not Transaction._running_on_start_hooks:
                Transaction._running_on_start_hooks = True
                try:
                    for hook in Transaction._on_start_hooks:
                        hook()
                finally:
                    Transaction._running_on_start_hooks = False
            Transaction._current_transaction = Transaction()

    def _prioritized(self, rank: Node, action: Handler["Transaction"]) -> None:
        entry = Entry(rank, action)
        heappush(self._prioritized_q, entry)

    def last(self, action: Callable[[], None]) -> None:
        """ Add an action to run after all :meth:`_prioritized` actions. """
        self._last_q.append(action)

    def _post(self, child_idx: int, action: Handler["Transaction"]) -> None:
        """ Add an action to run after all :meth:`last` actions. """
        if self._post_q is None:
            self._post_q = {}
        # If an entry exists already, combine the old one with the new one.
        if child_idx in self._post_q:
            existing = self._post_q[child_idx]
            def new(trans: Transaction) -> None:
                existing(trans)
                action(trans)
            self._post_q[child_idx] = new
        else:
            self._post_q[child_idx] = action

    @staticmethod
    def post(action: Callable[[], None]) -> None:
        """
        Execute the specified code after the current transaction is closed,
        or immediately if there is no current transaction.
        """
        if Transaction._current_transaction is not None:
            with Transaction._post_q_lock:
                if not Transaction._current_transaction._closed:
                    Transaction._current_transaction._post(
                        -1, lambda _: action())
                    return
        Transaction.run(action)

    def _check_regen(self) -> None:
        """
        If the priority queue has entries in it when we modify any of
        the nodes' ranks, then we need to re-generate it to make sure
        it's up-to-date.
        """
        if self._to_regen:
            self._to_regen = False
            self._prioritized_q.sort()

    def close(self) -> None:
        """ Close this transaction. Runs all scheduled actions. """

        while True:
            self._check_regen()
            if len(self._prioritized_q) == 0:
                break
            entry = heappop(self._prioritized_q)
            entry.action(self)

        for action in self._last_q:
            action()
        self._last_q.clear()

        with Transaction._post_q_lock:
            self._closed = True
            if self._post_q is not None:
                # _post_q must be traversed in the order of the indices
                for index in sorted(self._post_q.keys()):
                    handler = self._post_q.pop(index)
                    parent_trans = Transaction._current_transaction
                    try:
                        if index >= 0:
                            trans = Transaction()
                            Transaction._current_transaction = trans
                            try:
                                handler(trans)
                            finally:
                                trans.close()
                        else:
                            Transaction._current_transaction = None
                            handler(None)
                    finally:
                        Transaction._current_transaction = parent_trans



from threading import RLock
import traceback
from typing import \
    Any, \
    Callable, \
    Generic, \
    Iterable, \
    List, \
    Optional, \
    Sequence, \
    Set, \
    Tuple, \
    TypeVar, \
    Union
import weakref

from sodiumfrp.lazy import Lazy
from sodiumfrp.listener import Listener
from sodiumfrp.node import NODE_NULL, Node, Target
from sodiumfrp.transaction import Transaction
from sodiumfrp.typing import Handler, TransactionHandler
from sodiumfrp.unit import Unit, UNIT

A = TypeVar("A")
B = TypeVar("B")
C = TypeVar("C")
S = TypeVar("S")
T = TypeVar("T")


class Stream(Generic[A]):
    """
    Represents a stream of discrete events/firings containing values of
    type `A`.
    """

    _keep_listeners_alive: Set[Listener] = set()
    _keep_listeners_alive_lock: RLock = RLock()

    def __init__(self,
            node: Node,
            finalizers: List[Listener],
            firings: List[A]) -> None:
        """ Shouldn't be constructed explicitly. """
        self._node = node
        self._finalizers = finalizers
        self._firings = firings

    @staticmethod
    def never() -> "StreamWithSend[A]":
        """ A stream that never fires. """
        return StreamWithSend()

    def listen(self, handler: Handler[A]) -> Listener:
        """
        Listen for events/firings on this stream. This is the observer
        pattern. The returned :class:`~sodiumfrp.listener.Listener` has an
        :meth:`~sodiumfrp.listener.Listener.unlisten` method
        to cause the listener to be removed. This is an **operational**
        mechanism is for interfacing between the world of I/O and for FRP.

        :param handler: The handler to execute when there's a new value.
            You should make no assumptions about what thread you are
            called on, and the handler should not block. You are not allowed
            to use :meth:`CellSink.send` or :meth:`StreamSink.send` in
            the handler. An exception will be thrown, because you are not
            meant to use this to create your own primitives.
        """
        l0 = self.listen_weak(handler)
        def unlisten(_self: Listener) -> None:
            l0.unlisten()
            with Stream._keep_listeners_alive_lock:
                Stream._keep_listeners_alive.remove(_self)
        l = Listener(unlisten)
        with Stream._keep_listeners_alive_lock:
            Stream._keep_listeners_alive.add(l)
        return l

    def _listen(self,
            target: Node,
            action: TransactionHandler[A]) -> Listener:
        return Transaction._apply(
            lambda trans1: self._listen_internal(
                target, trans1, action, False))

    def listen_weak(self, action: Handler[A]) -> Listener:
        """
        A variant of :meth:`listen` that will deregister the listener
        automatically if the listener is garbage collected. With
        :meth:`listen`, the listener is only deregistered if
        :meth:`Listener.unlisten() <sodiumfrp.listener.Listener.unlisten>`
        is called explicitly.

        This method should be used for listeners that are to be passed to
        :meth:`Stream.add_cleanup` to ensure that things don't get kept
        alive when they shouldn't.
        """
        return self._listen(NODE_NULL, lambda trans2, a: action(a))

    def _listen_internal(self,
            target: Node,
            trans: Transaction,
            action: TransactionHandler[A],
            suppress_earlier_firings: bool) -> Listener:
        node_target_: List[Target] = [None]
        with Transaction._listeners_lock:
            if self._node._link_to(action, target, node_target_):
                trans._to_regen = True
        node_target = node_target_[0]
        firings = self._firings.copy()
        if (not suppress_earlier_firings) and (len(firings) > 0):
            def handler(trans2: Optional[Transaction]) -> None:
                for firing in firings:
                    Transaction.in_callback += 1
                    try:
                        action(trans2, firing)
                    except:
                        # Don't allow transactions to interfere with Sodium
                        # internals.
                        _print_stack()
                    finally:
                        Transaction.in_callback -= 1
            trans._prioritized(target, handler)

        captures: List[Any] = [
            # It's essential that we keep the listener alive while
            # the caller holds the Listener, so that the finalizer doesn't
            # get triggered.
            self,
            # It's also essential that we keep the action alive, since
            # the node uses a weak reference.
            action,
            node_target
        ]

        def unlisten(_self: Listener) -> None:
            event, _, target = captures
            with Transaction._listeners_lock:
                if event is not None:
                    event._node._unlink_to(target)
                    # Release captured references
                    captures[0] = None
                    captures[1] = None
                    captures[2] = None

        return Listener(unlisten)

    def map(self, f: Callable[[A], B]) -> "Stream[B]":
        """
        Transform the stream's event values according to the supplied
        function, so the returned Stream's event values reflect the value
        of the function applied to the input Stream's event values.

        :param f: Function to apply to convert the values. It may construct
            FRP logic or use :meth:`Cell.sample` in which case
            it is equivalent to :meth:`Stream.snapshot`'ing the cell. Apart
            from this the function must be **referentially transparent**.
        """
        out: StreamWithSend[B] = StreamWithSend()
        l = self._listen(out._node,
            lambda trans2, a: out._send(trans2, f(a)))
        return out._unsafe_add_cleanup(l)

    def map_to(self, b: B) -> "Stream[B]":
        """
        Returns a stream that outputs `b` for each event of the source
        stream.
        """
        return self.map(lambda _: b)

    def starmap(self, f: Callable[...,T]) -> "Stream[T]":
        """
        Like :meth:`Stream.map`, except that the elements of the stream are
        iterables that are unpacked as arguments when passing to the mapping
        function.
        """
        return self.map(lambda t: f(*t))

    def hold(self, init_value: A) -> "Cell[A]":
        """
        Create a cell with the specified initial value, that is updated
        by this stream's event values.

        There is an implicit delay: state updates caused by event firings
        don't become visible as the cell's current value as viewed by
        :meth:`Stream.snapshot` until the following transaction.
        To put this another way, :meth:`Stream.snapshot` always sees
        the value of a cell as it was before any state changes from
        the current transaction.
        """
        assert isinstance(self, StreamWithSend)
        return Transaction._apply(
            lambda trans: Cell(self, init_value))

    def hold_lazy(self, init_value: Lazy[A]) -> "Cell[A]":
        """
        A variant of :meth:`hold` with an initial value captured
        by :meth:`Cell.sample_lazy`.
        """
        return Transaction._apply(
            lambda trans: self._hold_lazy(trans, init_value))

    def _hold_lazy(self, _: Transaction, init_value: Lazy[A]) -> "Cell[A]":
        return LazyCell(self, init_value)

    def snapshot(self, *args: Union["Cell", Callable[...,T]]) -> "Stream[T]":
        """
        Return a stream whose events are the result of the combination, using
        the specified function, of the input stream's event value and
        the values of the provided cells at the time of the event.

        The comibining function is optional and, if it is present, must be
        provided as the last argument. If the function is omitted,
        the default combining function will be used, which ignores
        the stream's value and returns the values of the provided cells
        as a tuple. If only one cell provided, the default combining function
        returns just the value of this cell.

        There is an implicit delay: state updates caused by event firings
        being held with :meth:`Stream.hold` don't become visible as the cell's
        current value until the following transaction. To put this another
        way, :meth:`Stream.snapshot` always sees the value of a cell as it
        was before any state changes from the current transaction.
        """

        def parse_args() -> Tuple[Tuple, Callable]:
            if len(args) == 0:
                raise RuntimeError(
                    "Stream.snapshot() expects one or more arguments")

            def check_cell_types(cells: Iterable[Cell]) -> None:
                for index, cell in enumerate(cells):
                    if not isinstance(cell, Cell):
                        raise TypeError(
                            f"Argument {index} must be of type Cell. "
                            f"Actual type: {type(cell)}")

            if callable(args[-1]):
                cells = args[:-1]
                if len(cells) == 0:
                    raise RuntimeError("No cells were provided")
                check_cell_types(cells)
                f = args[-1]
                combine = lambda a: \
                    f(a, *map(lambda c: c._sample_no_trans(), cells))
                return cells, combine
            else:
                cells = args
                check_cell_types(cells)
                if len(cells) == 1:
                    cell = cells[0]
                    combine = lambda a: cell._sample_no_trans()
                else:
                    combine = lambda a: \
                        tuple(map(lambda c: c._sample_no_trans(), cells))
                return cells, combine

        cells, combine = parse_args()

        out: StreamWithSend[T] = StreamWithSend()
        l = self._listen(out._node,
            lambda trans2, a: out._send(trans2, combine(a)))
        return out._unsafe_add_cleanup(l)

    def or_else(self, s: "Stream[A]") -> "Stream[A]":
        """
        Variant of :meth:`merge_with` that merges two streams and will drop
        an event in the simultaneous case.

        In the case where two events are simultaneous (i.e. both within
        the same transaction), the event from `this` will take precedence,
        and the event from `s` will be dropped. If you want to specify your
        own combining function, use :meth:`merge_with`. `s1.or_else(s2)` is
        equivalent to `s1.merge_with(s2, lambda l, r: l)`.

        The name `or_else` is used instead of `merge_with` to make it
        really clear that care should be taken, because events can be dropped.
        """
        return self.merge_with(s, lambda left, right: left)

    @staticmethod
    def _merge(ea: "Stream[A]", eb: "Stream[A]") -> "Stream[A]":
        out: StreamWithSend[A] = StreamWithSend()
        left = Node(0)
        right = out._node
        node_target_: List[Target] = [None]
        left._link_to(lambda _: None, right, node_target_)
        node_target = node_target_[0]
        h = out._send
        l1 = ea._listen(left, h)
        l2 = eb._listen(right, h)
        def unlisten(_self: Listener) -> None:
            left._unlink_to(node_target)
        return out \
            ._unsafe_add_cleanup(l1) \
            ._unsafe_add_cleanup(l2) \
            ._unsafe_add_cleanup(Listener(unlisten))

    def merge_with(self,
            s: "Stream[A]",
            f: Callable[[A, A], A]) -> "Stream[A]":
        """
        Merge two streams of the same type into one, so that events on
        either input appear on the returned stream.

        If the events are simultaneous (that is, one event from this and one
        from `s` occurring in the same transaction), combine them into one
        using the specified combining function so that the returned stream
        is guaranteed only ever to have one event per transaction. The event
        from `self` will appear at the left input of the combining
        function, and the event from `s` will appear at the right.

        :param f: Function to combine the values. It may construct FRP logic
            or use :meth:`Cell.sample`. Apart from this the function must be
            **referentially transparent**.
        """
        return Transaction._apply(
            lambda trans: Stream._merge(self, s)._coalesce(trans, f))

    @staticmethod
    def or_else_(*streams: "Stream[A]") -> "Stream[A]":
        """
        Variant of :meth:`or_else` that merges a collection of streams.
        """
        return Stream.merge(lambda left, right: left, *streams)

    @staticmethod
    def merge(
            f: Callable[[A,A],A],
            *streams: "Stream[A]") -> "Stream[A]":
        """
        Variant of :meth:`merge_with` that merges multiple streams.
        """
        return Stream._merge_many(streams, 0, len(streams), f)

    @staticmethod
    def _merge_many(
            streams: Sequence["Stream[A]"],
            start: int,
            end: int,
            f: Callable[[A,A],A]) -> "Stream[A]":
        length = end - start
        if length == 0:
            return Stream.never()
        elif length == 1:
            return streams[start]
        elif length == 2:
            left = streams[start]
            right = streams[start + 1]
            return left.merge_with(right, f)
        else:
            mid = (start + end) // 2
            left = Stream._merge_many(streams, start, mid, f)
            right = Stream._merge_many(streams, mid, end, f)
            return left.merge_with(right, f)

    def _coalesce(self,
            trans1: Transaction,
            f: Callable[[A,A],A]) -> "Stream[A]":
        out: StreamWithSend[A]  = StreamWithSend()
        h = CoalesceHandler(f, weakref.ref(out))
        l = self._listen_internal(out._node, trans1, h, False)
        return out._unsafe_add_cleanup(l)

    def _last_firing_only(self, trans: Transaction) -> "Stream[A]":
        """
        Clean up the output by discarding any firing other than the last one.
        """
        return self._coalesce(trans, lambda first, second: second)

    def filter(self, predicate: Callable[[A], bool]) -> "Stream[A]":
        """
        Return a stream that only outputs events for which the predicate
        returns `True`.
        """
        out: StreamWithSend[A] = StreamWithSend()
        def handler(trans2: Transaction, a: A) -> None:
            if predicate(a):
                out._send(trans2, a)
        l = self._listen(out._node, handler)
        return out._unsafe_add_cleanup(l)

    def gate(self, c: "Cell[bool]") -> "Stream[A]":
        """
        Return a stream that only outputs events from the input stream
        when the specified cell's value is `True`.
        """
        skip = object()
        return self \
            .snapshot(c, lambda a, pred: a if pred else skip) \
            .filter(lambda x: x is not skip)

    def collect(self,
            init_state: S,
            f: Callable[[A, S], Tuple[B, S]]) -> "Stream[B]":
        """
        Transform an event with a generalized state loop (a Mealy machine).
        The function is passed the input and the old state and returns
        the new state and output value.

        :param f: Function to apply to update the state. It may construct FRP
            logic or use :meth:`Cell.sample` in which case it is equivalent
            to :meth:`Stream.snapshot`'ing the cell. Apart from this
            the function must be **referentially transparent**.
        """
        return self.collect_lazy(Lazy(lambda: init_state), f)

    def collect_lazy(self,
            init_state: Lazy[S],
            f: Callable[[A, S], Tuple[B, S]]) -> "Stream[B]":
        """
        A variant of :meth:`collect` that takes an initial state returned by
        :meth:`Cell.sample_lazy`.
        """
        def handler() -> Stream[B]:
            ea = self
            es: StreamLoop[S] = StreamLoop()
            s = es.hold_lazy(init_state)
            ebs = ea.snapshot(s, f)
            eb = ebs.map(lambda bs: bs[0])
            es_out = ebs.map(lambda bs: bs[1])
            es.loop(es_out)
            return eb

        return Transaction.run(handler)

    def accum(self, init_state: S, f: Callable[[A, S], S]) -> "Cell[S]":
        """
        Accumulate on input event, outputting the new state each time.

        :param f: Function to apply to update the state. It may construct FRP
            logic or use :meth:`Cell.sample` in which case it is equivalent
            to :meth:`Stream.snapshot`'ing the cell. Apart from this
            the function must be **referentially transparent**.
        """
        return self.accum_lazy(Lazy(lambda: init_state), f)

    def accum_lazy(self,
            init_state: Lazy[S],
            f: Callable[[A, S], S]) -> "Cell[S]":
        """
        A variant of :meth:`accum` that takes an initial state returned by
        :meth:`Cell.sample_lazy`.
        """
        def handler() -> Cell[S]:
            ea = self
            es: StreamLoop[S] = StreamLoop()
            s = es.hold_lazy(init_state)
            es_out = ea.snapshot(s, f)
            es.loop(es_out)
            return es_out.hold_lazy(init_state)

        return Transaction.run(handler)

    def once(self) -> "Stream[A]":
        """
        Return a stream that outputs only one value: the next event of the
        input stream, starting from the transaction in which `once()` was
        invoked.
        """
        # This is a bit long-winded but it's efficient because it deregisters
        # the listener.
        la: List[Optional[Listener]] = [None]
        out: StreamWithSend[A] = StreamWithSend()

        def handler(trans: Transaction, a: A) -> None:
            if la[0] is not None:
                out._send(trans, a)
                la[0].unlisten()
                la[0] = None

        la[0] = self._listen(out._node, handler)
        return out._unsafe_add_cleanup(la[0])

    def calm(self) -> "Stream[A]":
        """
        Returns a stream that ignores repeating values produced
        by the source stream.
        """
        return self._calm(None)

    def _calm(self, init) -> "Stream[A]":
        reducer = lambda new, old: (None, old) if new == old else (new, new)
        return self.collect(init, reducer).filter(lambda x: x is not None)

    def _unsafe_add_cleanup(self, cleanup: Listener) -> "Stream[A]":
        """
        This is not thread-safe, so one of these two conditions must apply:
        1. We are within a transaction, since in the current implementation
           a transaction locks out all other threads.
        2. The object on which this is being called has not yet been
           returned from the method where it was created, so it can't
           be shared between threads.
        """
        self._finalizers.append(cleanup)
        return self

    def add_cleanup(self, cleanup: Listener) -> "Stream[A]":
        """
        Attach a listener to this stream so that its
        :meth:`Listener.unlisten() <sodiumfrp.listener.Listener.unlisten>`
        is invoked when this stream is garbage collected. Useful for
        functions that initiate I/O, returning the result of it through
        a stream.

        You must use this only with listeners returned
        by :meth:`listen_weak()` so that things don't get kept alive when
        they shouldn't.
        """
        def helper() -> Stream[A]:
            fs_new = self._finalizers.copy()
            fs_new.append(cleanup)
            return Stream(self._node, fs_new, self._firings)

        return Transaction.run(helper)

    def __del__(self) -> None:
        for l in self._finalizers:
            l.unlisten()


class StreamWithSend(Stream[A]):

    def __init__(self) -> None:
        super().__init__(Node(0), [], [])

    def _send(self, trans: Transaction, a: A) -> None:
        if len(self._firings) == 0:
            trans.last(lambda: self._firings.clear())
        self._firings.append(a)
        with Transaction._listeners_lock:
            listeners = set(self._node._listeners)
        for target in listeners:
            def handler(
                    trans2: Transaction,
                    # Pass target as default argument, as loop variables
                    # shouldn't be captured by a closure
                    target: Target = target) -> None:
                Transaction.in_callback += 1
                try:
                    uta = target.action() # Dereference the weak reference
                    if uta is not None: # If it hasn't been GC'ed, call it
                        uta(trans2, a)
                except:
                    # Don't allow transactions to interfere with Sodium
                    # internals.
                    _print_stack()
                finally:
                    Transaction.in_callback -= 1
            trans._prioritized(target.node, handler)


class StreamSink(StreamWithSend[A]):
    """
    A stream that allows values to be pushed into it, acting as an interface
    between the world of I/O and the world of FRP. Code that exports
    `StreamSinks` for read-only use should downcast to :class:`Stream`.
    """

    def __init__(self, f: Callable[[A, A], A] = None) -> None:
        """
        Construct a `StreamSink`. Use the provided function to combine values,
        that were sent to the stream during the same transaction, into a
        single event. If the function is `None`, :meth:`send` throws
        an exception, if it is called more then once per transaction.

        The combining function should be **associative**.

        :param f: Function to combine the values. It may construct FRP logic
            or use :meth:`Cell.sample`. Apart from this the function
            must be **referentially transparent**.
        """
        super().__init__()
        def error(left: A, right: A) -> A:
            raise RuntimeError("send() called more than once per "
                "transaction, which isn't allowed. Did you want to combine "
                "the events? Then pass a combining function to your "
                "StreamSink constructor.");
        self._coalescer = CoalesceHandler(
            f if f is not None else error, weakref.ref(self))

    def send(self, a: A) -> None:
        """
        Send a value to be made available to consumers of the stream.
        `send()` may not be used inside handlers registered with
        :meth:`Stream.listen` or :meth:`Cell.listen`. An exception
        will be thrown, because `StreamSink` is for interfacing I/O
        to FRP only. You are not meant to use this to define your own
        primitives.

        :param a: value to push into the cell.
        """
        def handler(trans: Transaction) -> None:
            if trans.in_callback > 0:
                raise RuntimeError("You are not allowed to use send() "
                    "inside a Sodium callback");
            self._coalescer(trans, a)
        Transaction._run_handler(handler)


class StreamLoop(StreamWithSend[A]):
    """
    A forward reference for a `Stream` equivalent to the `Stream` that
    is referenced.
    """

    def __init__(self) -> None:
        """ Make an instance of `StreamLoop`. """
        super().__init__()
        self._assigned = False
        if Transaction.get_current_transaction() is None:
            raise RuntimeError("StreamLoop/CellLoop must be used within "
                "an explicit transaction")

    def loop(self, ea_out: Stream[A]) -> None:
        """
        Resolve the loop to specify what the `StreamLoop` was a forward
        reference to. It must be invoked inside the same transaction as
        the place where the `StreamLoop` is used. This requires you to create
        an explicit transaction with :meth:`Transaction.run()
        <sodiumfrp.transaction.Transaction.run>`.
        """
        if self._assigned:
            raise RuntimeError("StreamLoop looped more than once")
        self._assigned = True
        Transaction.run(
            lambda: self._unsafe_add_cleanup(
                ea_out._listen(self._node, self._send)
            )
        )


class CoalesceHandler(Generic[A]):

    def __init__(self,
            f: Callable[[A,A], A],
            # Pass a weak reference here, to prevent circular references
            # in Stream._coalesce() and StreamSink.__init__().
            out: "weakref.ReferenceType[StreamWithSend[A]]") -> None:
        self._f = f
        self._out = out
        self._accum_valid = False
        self._accum: A = None

    def __call__(self, trans1: Transaction, a: A) -> None:
        if self._accum_valid:
            self._accum = self._f(self._accum, a)
        else:
            def handler(trans2: Optional[Transaction]) -> None:
                self._out()._send(trans2, self._accum)
                self._accum_valid = False
                self._accum = None
            trans1._prioritized(self._out()._node, handler)
            self._accum = a
            self._accum_valid = True


class Cell(Generic[A]):
    """ Represents a value of type `A` that changes over time. """

    @staticmethod
    def constant(value: A) -> "Cell[A]":
        """ Returns a cell that never changes, always holding the `value`. """
        return Cell(Stream.never(), value)

    def __init__(self, stream: StreamWithSend[A], init_value: A) -> None:
        """ Shouldn't be constructed explicitly. """
        self._stream = stream
        self._value = init_value
        self._value_update: A = None
        self._cleanup: Listener = None
        self._lazy_init_value: Lazy[A] = None # Use by LazyCell
        def handler(trans1: Transaction) -> None:
            def handler2(trans2: Transaction, a: A) -> None:
                if self._value_update is None:
                    def run() -> None:
                        self._value = self._value_update
                        self._lazy_init_value = None
                        self._value_update = None
                    trans2.last(run)
                self._value_update = a
            self._cleanup = stream._listen_internal(
                NODE_NULL, trans1, handler2, False)
        Transaction._run_handler(handler)

    def sample(self) -> A:
        """
        Sample the cell's current value.

        It may be used inside the functions passed to primitives that apply
        them to :class:`Stream` including :meth:`Stream.map` (in which case
        it is equivalent to snapshotting the cell), :meth:`Stream.snapshot`,
        :meth:`Stream.filter` and :meth:`Stream.merge`.

        It should generally be avoided in favour of :meth:`listen` so you
        don't miss any updates, but in many circumstances it makes sense.
        """
        return Transaction._apply(lambda trans: self._sample_no_trans())

    def sample_lazy(self) -> Lazy[A]:
        """
        A variant of :meth:`sample` that works with :class:`CellLoop` when
        they haven't been looped yet.  It should be used in any code that's
        general enough that it could be passed a :class:`CellLoop`.
        """
        return Transaction._apply(lambda trans: self._sample_lazy(trans))

    def _sample_lazy(self, trans: Transaction) -> Lazy[A]:

        class LazySample:
            def __init__(self,
                    cell: Cell[A],
                    has_value: bool,
                    value: A) -> None:
                self.cell = cell
                self.has_value = has_value
                self.value = value

        s = LazySample(cell=self, has_value=None, value=None)
        def handler() -> None:
            s.value = self._value_update \
                if self._value_update is not None \
                else self._sample_no_trans()
            s.has_value = True
            s.cell = None
        trans.last(handler)
        return Lazy(lambda: s.value if s.has_value else s.cell.sample())

    def _sample_no_trans(self) -> A:
        return self._value

    def _updates(self) -> Stream[A]:
        return self._stream

    def _value_stream(self, trans1: Transaction) -> Stream[A]:
        s_spark: StreamWithSend[Unit] = StreamWithSend()
        trans1._prioritized(
            s_spark._node, lambda trans2: s_spark._send(trans2, UNIT))
        s_initial: Stream[A] = s_spark.snapshot(self)
        return s_initial.merge_with(
            self._updates(), lambda left, right: right)

    def map(self, f: Callable[[A], B]) -> "Cell[B]":
        """
        Transform the cell's value according to the supplied function,
        so the returned cell always reflects the value of the function
        applied to the input cell's value.
        """
        return Transaction._apply(
            lambda trans: self \
                ._updates() \
                .map(f) \
                ._hold_lazy(trans, self._sample_lazy(trans).lift(f))
        )

    def starmap(self, f: Callable[...,T]) -> "Cell[T]":
        """
        Like :meth:`Cell.map`, except that the value of the cell is
        an iterable that is unpacked as arguments when passing to
        the mapping function.
        """
        return self.map(lambda t: f(*t))

    # TODO lift over multiple cells at once
    def lift(self, b: "Cell[B]", f: Callable[[A,B],C]) -> "Cell[C]":
        """
        Lift a binary function into cells, so the returned cell always
        reflects the specified function applied to the input cells' values.

        :param f: Function to apply. It must be **referentially transparent**.
        """
        bf = self.map(lambda aa: lambda bb: f(aa, bb))
        return Cell.apply(bf, b)

    @staticmethod
    def apply(bf: "Cell[Callable[[A],B]]", ba: "Cell[A]") -> "Cell[B]":
        """
        Apply a value inside a cell to a function inside a cell. This is the
        primitive for all function lifting.
        """
        def handler(trans0: Transaction) -> "Cell[B]":
            out: StreamWithSend[B] = StreamWithSend()

            class ApplyHandler:
                def __init__(self) -> None:
                    self.f: Callable[[A],B] = None
                    self.f_present = False
                    self.a: A = None
                    self.a_present = False
                def run(self, trans1: Transaction) -> None:
                    trans1._prioritized(out._node,
                        lambda trans2: out._send(trans2, self.f(self.a)))

            out_target = out._node
            in_target = Node(0)
            node_target_: List[Target] = [None]
            in_target._link_to(lambda _: None, out_target, node_target_)
            node_target = node_target_[0]
            h = ApplyHandler()

            def handler_f(trans1: Transaction, f: Callable[[A],B]) -> None:
                h.f = f
                h.f_present = True
                if h.a_present:
                    h.run(trans1)

            l1 = bf._value_stream(trans0)._listen(in_target, handler_f)

            def handler_a(trans1: Transaction, a: A) -> None:
                h.a = a
                h.a_present = True
                if h.f_present:
                    h.run(trans1)

            l2 = ba._value_stream(trans0)._listen(in_target, handler_a)

            return out \
                ._last_firing_only(trans0) \
                ._unsafe_add_cleanup(l1) \
                ._unsafe_add_cleanup(l2) \
                ._unsafe_add_cleanup(
                    Listener(lambda _: in_target._unlink_to(node_target))) \
                .hold_lazy(
                    Lazy(
                        lambda: bf._sample_no_trans()(ba._sample_no_trans())
                    )
                )

        return Transaction._apply(handler)

    def switch(self) -> "Cell":
        """
        Do either :meth:`switch_stream` or :meth:`switch_cell` depending on
        the type of the value of the cell.
        """
        if isinstance(self._value, Stream):
            return Cell.switch_stream(self)
        elif isinstance(self._value, Cell):
            return Cell.switch_cell(self)
        else:
            raise TypeError(
                "Can't apply Cell.switch() to a Cell holding value of type "
                f"{type(self._value)}. Type of the value must be either "
                "Stream or Cell.")

    @staticmethod
    def switch_cell(bba: "Cell[Cell[A]]") -> "Cell[A]":
        """
        Unwrap a cell inside another cell to give a time-varying cell
        implementation.
        """
        def helper(trans0: Transaction) -> "Cell[A]":
            za = bba.sample_lazy().lift(lambda ba: ba.sample())
            out: StreamWithSend[A] = StreamWithSend()

            class _Handler:

                def __init__(self) -> None:
                    self.current_listener: Listener = None

                def __call__(self, trans2: Transaction, ba: Cell[A]) -> None:
                    # Note: If any switch takes place during a transaction,
                    # then the _last_firing_only() below will always cause
                    # a sample to be fetched from the one we just switched
                    # to. So anything from the old input cell that might
                    # have happened during this transaction will be suppressed.
                    if self.current_listener is not None:
                        self.current_listener.unlisten()
                    self.current_listener = ba \
                        ._value_stream(trans2) \
                        ._listen_internal(out._node, trans2, out._send, False)

                def __del__(self) -> None:
                    if self.current_listener is not None:
                        self.current_listener.unlisten()

            l1 = bba._value_stream(trans0)._listen(out._node, _Handler())
            return out \
                ._last_firing_only(trans0) \
                ._unsafe_add_cleanup(l1) \
                .hold_lazy(za)

        return Transaction._apply(helper)

    @staticmethod
    def switch_stream(bea: "Cell[Stream[A]]") -> Stream[A]:
        """
        Unwrap a stream inside a cell to give a time-varying stream
        implementation.
        """
        return Transaction._apply(
            lambda trans: Cell._switch_stream(trans, bea))

    @staticmethod
    def _switch_stream(
            trans1: Transaction,
            bea: "Cell[Stream[A]]") -> Stream[A]:
        out: StreamWithSend[A] = StreamWithSend()
        h2 = out._send

        class _Handler:

            def __init__(self) -> None:
                self.current_listener = bea \
                    ._sample_no_trans() \
                    ._listen_internal(out._node, trans1, h2, False)

            def __call__(self, trans2: Transaction, ea: Stream[A]) -> None:
                def run() -> None:
                    if self.current_listener is not None:
                        self.current_listener.unlisten()
                    self.current_listener = ea._listen_internal(
                        out._node, trans2, h2, True)
                trans2.last(run)

            def __del__(self) -> None:
                if self.current_listener is not None:
                    self.current_listener.unlisten()

        h1 = _Handler()
        l1 = bea._updates()._listen_internal(out._node, trans1, h1, False)
        return out._unsafe_add_cleanup(l1)

    def calm(self) -> "Cell[A]":
        """
        Returns a cell that holds the same value as the source cell, but
        ignores updates that don't change value of the cell.
        """
        init = self.sample()
        return self._updates()._calm(init).hold(init)

    def __del__(self) -> None:
        if self._cleanup is not None:
            self._cleanup.unlisten()

    def listen(self, action: Callable[[A],None]) -> Listener:
        """
        Listen for updates to the value of this cell. This is the observer
        pattern. The returned :class:`~sodiumfrp.listener.Listener` has
        an :meth:`~sodiumfrp.listener.Listener.unlisten` method
        to cause the listener to be removed. This is an **operational**
        mechanism for interfacing between the world of I/O and for FRP.

        :param action: the handler to execute when there's a new value.
            You should make no assumptions about what thread you are called
            on, and the handler should not block. You are not allowed
            to use :meth:`CellSink.send` or :meth:`StreamSink.send` in
            the handler. An exception will be thrown, because you are not
            meant to use this to create your own primitives.
        """
        return Transaction._apply(
            lambda trans: self._value_stream(trans).listen(action))


class CellSink(Cell[A]):
    """
    A cell that allows values to be pushed into it, acting as an interface
    between the world of I/O and the world of FRP. Code that exports
    `CellSinks` for read-only use should downcast to
    :class:`~sodiumfrp.primitives.Cell`.
    """

    def __init__(self, init_value: A, f: Callable[[A,A],A] = None) -> None:
        """
        Construct a writable cell with the specified initial value.
        If multiple values are sent in the same transaction, the specified
        function is used to combine them. If the function isn't provided,
        :meth:`send` throws an exception when called multiple times from the
        same transaction.
        """
        super().__init__(StreamSink(f), init_value)

    def send(self, a: A) -> None:
        """
        Send a value, modifying the value of the cell. `send()` may not be
        used inside handlers registered with :meth:`Stream.listen` or
        :meth:`Cell.listen`. An exception will be thrown, because `CellSink`
        is for interfacing I/O to FRP only. You are not meant to use this
        to define your own primitives.

        :param a: value to push into the cell.
        """
        self._stream.send(a)


class LazyCell(Cell[A]):

    def __init__(self, event: Stream[A], lazy_init_value: Lazy[A]) -> None:
        super().__init__(event, None)
        self._lazy_init_value = lazy_init_value

    def _sample_no_trans(self) -> A:
        if (self._value is None) and (self._lazy_init_value is not None):
            self._value = self._lazy_init_value.get()
            self._lazy_init_value = None
        return self._value


class CellLoop(LazyCell[A]):
    """
    A forward reference for a :class:`Cell` equivalent to the `Cell` that is
    referenced.
    """

    def __init__(self) -> None:
        """ Make an instance of `CellLoop`. """
        super().__init__(StreamLoop(), None)

    def loop(self, a_out: Cell[A]) -> None:
        """
        Resolve the loop to specify what the `CellLoop` was a forward
        reference to. It must be invoked inside the same transaction as
        the place where the `CellLoop` is used. This requires you to create
        an explicit transaction with
        :meth:`Transaction.run() <sodiumfrp.transaction.Transaction.run>`.
        """
        def handler(trans: Transaction) -> Unit:
            self._stream.loop(a_out._updates())
            self._lazy_init_value = a_out._sample_lazy(trans)
            return UNIT
        Transaction._apply(handler)

    def _sample_no_trans(self) -> A:
        if not self._stream._assigned:
            raise RuntimeError("CellLoop sampled before it was looped")
        return super()._sample_no_trans()

def _print_stack() -> None:
    """ Print full stack trace. format_exc() shows only part of it. """
    stack = traceback.format_stack()[:-2]
    exc = traceback.format_exc()
    stack_lines = "".join(stack).split("\n")[:-1]
    exc_lines = exc.split("\n")
    trace = "\n".join(exc_lines[:1] + stack_lines + exc_lines[1:])
    print(trace)



""" Utilities for working with continuous time. """

from collections import namedtuple
from threading import Condition, Event, Thread
import time
from typing import Callable, List, Optional, Set, Tuple, TypeVar

from sodiumfrp import Cell, CellSink, Stream, StreamSink, Transaction

A = TypeVar("A")
B = TypeVar("B")

Timer = namedtuple("Timer", ["time", "streams"])

class TimerSystem:

    def __init__(self) -> None:
        self._time_ms: CellSink[int] = CellSink(int(_now_ms()))
        self._timers: List[Timer] = []
        self._timers_changed = Condition()
        self._stop = Event()
        self._released = False

        worker = Thread(target=self._thread)
        worker.daemon = True
        worker.start()

        self._unregister_hook = Transaction.on_start(
            self._on_transaction_start)

    def __del__(self) -> None:
        if not self._released:
            self.release()

    def __enter__(self) -> "TimerSystem":
        return self

    def __exit__(self, *args) -> None:
        self.release()

    def release(self) -> None:
        self._released = True
        self._stop.set()
        self._unregister_hook()
        with self._timers_changed:
            # Unlock the thread and let it exit the infinite loop
            self._timers_changed.notify_all()

    def _thread(self) -> None:
        """
        Timers are executed at the beginning of a transaction. If there are
        not transactions, no timers will be executed. Thus, we need
        a separate thread that would generate transactions firing timers.
        """
        while not self._stop.is_set():
            with self._timers_changed:
                timeout = None
                while self._timers:
                    earliest_timer = self._timers[0]
                    now = _now_ms()
                    if earliest_timer.time > now:
                        timeout = (earliest_timer.time - now) * 0.001
                        break
                    else:
                        # If the time of the earliest timer has passed,
                        # start an empty transaction to trigger all passed
                        # timers through the transaction start hook
                        Transaction.run(lambda: None)
                self._timers_changed.wait(timeout)

    def _on_transaction_start(self) -> None:
        """
        This should be used as a :meth:`Transaction.on_start()
        <sodiumfrp.transaction.Transaction.start>` hook. At the
        beginning of each transaction, update `self._time_ms` and execute
        all timers that should've been fired by this time.
        """
        trans_time = int(_now_ms())
        time_cell_updated = False
        with self._timers_changed:
            while self._timers:
                timer = self._timers[0]
                timer_time = timer.time
                if timer_time > trans_time:
                    break
                # Update time and execute timers inside the same transaction
                # to guarantee simultaneity of events
                def timer_transaction() -> None:
                    self._time_ms.send(timer_time)
                    for stream in timer.streams:
                        stream.send(timer_time)
                    del self._timers[0]
                    # Don't need to notify _thread(). It will wake up and
                    # skip to the next available timer on its own.
                Transaction.run(timer_transaction)
                if timer_time == trans_time:
                    time_cell_updated = True
            if not time_cell_updated:
                self._time_ms.send(trans_time)

    def time_ms(self) -> Cell[int]:
        """
        Returns a cell that represents current time (in milliseconds).
        """
        return self._time_ms

    def at(self, alarm_time: Cell[Optional[int]]) -> Stream[int]:
        """
        Returns a stream that fires events at the points in time indicated
        by the value of the input cell. If the value of the cell is `None`
        or less than the current time, then no events will be fired and
        the timer set by the previous value of the cell will be canceled.
        """
        alarm_stream: StreamSink[int] = StreamSink()
        current_alarm_time_value: List[Optional[int]] = [None]
        def handler(alarm_time_value: Optional[int]) -> None:
            if current_alarm_time_value[0] is not None:
                if alarm_time_value != current_alarm_time_value[0]:
                    self._cancel_timer(
                        current_alarm_time_value[0], alarm_stream)
            if alarm_time_value is not None:
                if alarm_time_value > self._time_ms.sample():
                    self._set_timer(alarm_time_value, alarm_stream)
                    current_alarm_time_value[0] = alarm_time_value
                else:
                    current_alarm_time_value[0] = None
            else:
                current_alarm_time_value[0] = None
        listener = alarm_time.listen(handler)
        return alarm_stream.add_cleanup(listener)

    def _cancel_timer(self, time_: int, stream: StreamSink[int]) -> None:
        with self._timers_changed:
            timer, index = _bisect(self._timers, time_, lambda t: t.time)
            if timer is not None:
                timer.streams.discard(stream)
                if len(timer.streams) == 0:
                    del self._timers[index]
                    self._timers_changed.notify_all()

    def _set_timer(self, time_: int, stream: StreamSink[int]) -> None:
        with self._timers_changed:
            timer, index = _bisect(self._timers, time_, lambda t: t.time)
            if timer is not None:
                timer.streams.add(stream)
            else:
                self._timers.insert(index, Timer(time_, {stream}))
                self._timers_changed.notify_all()

def _now_ms() -> float:
    return time.monotonic() * 1000

def _bisect(
        a: List[A],
        x: B, key: Callable[[A],B]) -> Tuple[Optional[A], int]:
    length = len(a)
    lo = 0
    hi = length
    while lo < hi:
        mid = (lo + hi) // 2
        if key(a[mid]) < x:
            lo = mid + 1
        else:
            hi = mid
    index = lo
    if index < length and key(a[index]) == x:
        return a[index], index
    else:
        return None, index
